#    Author: Alex Savatieiev (osavatieiev@mirantis.com; a.savex@gmail.com)
#    Copyright 2019-2022 Mirantis, Inc.
import re

from datetime import datetime, timezone
from multiprocessing.dummy import Pool

from cfg_checker.common import logger_cli
from cfg_checker.common.exception import KubeException

from cfg_checker.helpers.console_utils import Progress
from cfg_checker.nodes import KubeNodes


_datetime_fmt = "%Y-%m-%d-%H-%M-%S.%f"

# parsers = [
#   [
#       <split lines_parser>,
#       <split data for a line parser>
#   ]
# ]
log_line_regex_parsers = [
    {
        # (all others) YYYY-MM-DD hh:mm:ss,nnn
        # 
        "lines": "\n(?=\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\,\d{3})",
        "items": ["date", "date", "type", "message"],
        "date_fmt": "%Y-%m-%d %H:%M:%S,%f"
    },
    {
        # 'stacklight/alerta-5fc6f5dfd-jj7ml'
        # '2022/04/01 23:12:37 [info] 26#26: *124156 client 127.0.0.1 closed keepalive connection\n'
        "lines": "\n(?=\d{4}\/\d{2}\/\d{2} \d{2}:\d{2}:\d{2}\,\d{3})",
        "items": ["date", "date", "type", "message"],
        "date_fmt": "%Y/%m/%d %H:%M:%S,%f"
    },
    {
        # (nova-api-osapi-....) YYYY-MM-DD hh:mm:ss,nnnnnn
        # '2022-04-01 23:08:11.806062 capabilities. Old policies are deprecated and silently going to be ignored'
        "lines": "\n(?=\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{6})",
        "items": ["date", "date", "message"],
        "date_fmt": "%Y-%m-%d %H:%M:%S.%f"
    },
    {
        # (nova-api-metadata..) nnn.nnn.nnn.nnn - - [DD/MMM/YYYY:hh:mm:ss +nnnn]
        # '172.16.35.67 - - [01/Apr/2022:22:23:14 +0000] "GET / HTTP/1.1" 200 98 1345 "-" "kube-probe/1.20+"'
        "lines": "\n(?=\d{1,3}.\d{1,3}.\d{1,3}.\d{1,3}.+-.+-.+\[\d{1,2}\/\S{3}\/\d{4}:\d{2}:\d{2}:\d{2}.\+\d{4}\])",
        "items": ["ip", None, None, "date", "date", "message"],
        "date_fmt": "[%d/%b/%Y:%H:%M:%S %z]"
    },
    {
        # (iam-proxy-alerta...) nnn.nnn.nnn.nnn:ppppp - uuid - - [YYYY/MM/DD hh:mm:ss]
        # '172.16.35.69:35298 - 9b68130c-4c3b-4abd-bb04-7ff5329ad644 - - [2022/04/01 23:00:50] 10.233.118.232:4180 GET - "/ping" HTTP/1.1 "kube-probe/1.20+" 200 2 0.000'
        "lines": "\n(?=\d{1,3}.\d{1,3}.\d{1,3}.\d{1,3}:\d{1,5}.+-.+\S{8}-\S{4}-\S{4}-\S{4}-\S{12}.+-.+-.+\[\d{4}\/\d{2}\/\d{2} \d{2}:\d{2}:\d{2}])",
        "items": ["ip", None, "uuid", None, None, "date", "date", "message"],
        "date_fmt": "[%Y/%m/%d %H:%M:%S]"
    }
]


class MosLogger(object):
    def __init__(
        self,
        config
    ):
        self.env_config = config
        return


class KubeMosLogger(MosLogger):
    def __init__(self, config):
        self.master = KubeNodes(config)
        super(KubeMosLogger, self).__init__(config)
        # Init ceph tools pod
        self.logs = {}
        self.merged_logs = {}

    def _keywords(self, tstring, keywords):
        return [True if k in tstring else False for k in keywords]

    def prepare_pods(self, ns_list, kw_list, inclusive_filter=True):
        logger_cli.info(
            "# Preparing pods, ns: {}; keywords: {}".format(
                ", ".join(ns_list) if ns_list else "*",
                ", ".join(kw_list) if kw_list else "*"
            )
        )
        # [ns, pod_name]
        _target_pods = []
        # validate ns
        _all_namespaces = self.master.list_namespaces()
        for _ns in _all_namespaces:
            if _ns in ns_list or not ns_list:
                _tmp_pods = []
                logger_cli.info("-> processing namespace '{}'".format(_ns))
                # list pods using mask
                logger_cli.debug("... getting pods list")
                _pods = self.master.list_resource_names_by_type_and_ns(
                    "pod",
                    ns=_ns
                )
                logger_cli.debug("... {} total pods found".format(len(_pods)))
                for _pod in _pods:
                    # _pod[_ns, _name]
                    _kw = self._keywords(_pod[1], kw_list) \
                        if kw_list else [True]
                    if not inclusive_filter and all(_kw):
                        _tmp_pods += [_pod]
                        logger_cli.debug(
                            "... added pod '{}'".format("/".join(_pod))
                        )
                    elif inclusive_filter and any(_kw):
                        _tmp_pods += [_pod]
                        logger_cli.debug(
                            "... added pod '{}'".format("/".join(_pod))
                        )
                    else:
                        # logger_cli.debug("... skipped pod '{}'".format(_pod))
                        pass
                logger_cli.info(
                    "-> {} of {} pods selected'".format(
                        len(_tmp_pods),
                        len(_pods)
                    )
                )
                _target_pods += _tmp_pods

        logger_cli.info("-> found {} pods total".format(len(_target_pods)))
        return _target_pods

    def _get_pod_log(self, params):
        def _detect_log_syntax(str_to_detect):
            _max_matched = 0
            _detected_pattern = None
            _detected_index = -1
            for idx in range(len(log_line_regex_parsers)):
                _pattern = log_line_regex_parsers[idx]
                _compiled = re.compile(_pattern["lines"])
                _pcount = len(_compiled.findall(str_to_detect))
                if _pcount > _max_matched:
                    _max_matched = _pcount
                    _detected_pattern = _compiled
                    _detected_index = idx
            if _detected_pattern:
                return _detected_index, _detected_pattern.split(str_to_detect)
            else:
                return _detected_index, None
        ns = params[0]
        name = params[1]
        _index, _log = _detect_log_syntax(
            self.master.get_logs_for_pod(name, ns)
        )
        return "{}/{}".format(ns, name), _index, _log

    def collect_logs(self, pods_list):

        # Prepare list of pods to collect
        # cmd = """
        # kubectl get pods -A -o=jsonpath='{range .items[*]}{.metadata.namespace}{"/"}{.metadata.name}{"\n"}{range .spec.containers[*]} {.name}{"\n"}{end}'
        # """

        # Inspect pods for containers
        # TODO: Handle containers

        logger_cli.info(
            "# Collecting logs using {} threads".format(
                self.env_config.threads
            )
        )
        # Do collect using pooled threads
        pool = Pool(self.env_config.threads)
        _params = []
        # Prepare params for getting logs
        for _ns, _pod_name in pods_list:
            _params.append([_ns, _pod_name])
        # Init progress bar
        _progress = Progress(len(_params))
        # Start pooled processing
        results = pool.imap_unordered(self._get_pod_log, _params)
        # Catch and parse results
        # [index, namepath, regex_index, lines]
        for _r in enumerate(results, start=1):
            _namepath = _r[1][0]
            _index = _r[1][1]
            _lines = _r[1][2]

            self.logs[_namepath] = {
                "parser_index": _index,
                "log": []
            }
            _total = len(_lines)

            # Filter lines
            # TODO: Filter logs using keywords, simple grep like
            _selected = _total

            # Save filtered
            self.logs[_namepath]["log"] = _lines
            # print progress
            _progress.write_progress(
                _r[0],
                note="'{}': {} lines loaded, {} filtered".format(
                    _namepath,
                    _total,
                    _selected
                )
            )

        _progress.end()
        pool.close()
        pool.join()

        return

    def merge_logs(self):
        logger_cli.info("# Merging logs")
        _merged = {}
        _parsed_count = 0
        _not_parsed_date__count = 0
        for _pod, _logs in self.logs.items():
            # get proper parser
            log_headers = log_line_regex_parsers[_logs["parser_index"]]["items"]
            for _line in _logs["log"]:
                # Prepare log entry
                log_line = {"namepath": _pod}
                # Split log into headers and tread remainder as one item
                item_list = _line.split()
                if not item_list:
                    continue
                header_count = len(log_headers)
                idx = 0
                while idx < header_count:
                    # Check if this is the last header
                    if idx == header_count - 1:
                        _item = " ".join(item_list)
                    else:
                        _item = item_list.pop(0)
                    _header = log_headers[idx]
                    if _header is None:
                        # Skip empty data
                        pass
                    elif _header in log_line:
                        # merge same headers
                        log_line[_header] = " ".join([log_line[_header], _item])
                    else:
                        log_line[_header] = _item

                    idx += 1

                # detect timestamp
                date_fmt = \
                    log_line_regex_parsers[_logs["parser_index"]]["date_fmt"]
                # DEBUG
                try:
                    _timestamp = datetime.strptime(log_line["date"], date_fmt)
                    # Since date is parsed, throw it away
                    _ = log_line.pop("date")
                    _parsed_count += 1
                except ValueError:
                    logger_cli.warning(
                        "WARNING ({}): Can't parse date '{}'"
                        " using '{}'".format(
                            log_line["namepath"],
                            log_line["date"],
                            date_fmt
                        )
                    )
                    # Use now as a timestamp
                    # TODO: Add failsafe using kube internal log date

                    _timestamp = datetime.now(timezone.utc)
                    _not_parsed_date__count += 1
                # _timestamp = _log_items[0]

                # and put undetected lines separatelly
                # save items using timestamps
                _merged[float(_timestamp.timestamp())] = log_line
        self.merged_logs = _merged
        logger_cli.info(
            "-> {} lines parsed, {} failed".format(
                _parsed_count,
                _not_parsed_date__count
            )
        )
        return

    def save_logs(self, filename):
        logger_cli.info("# Writing output file: '{}'".format(filename))
        with open(filename, 'w+') as ff:
            _log_iter = sorted(
                self.merged_logs.items(), key=lambda item: item[0]
            )
            for k, v in _log_iter:
                ff.write(
                    "{} {}: {}\n".format(
                        v.pop("namepath"),
                        datetime.fromtimestamp(k).strftime(_datetime_fmt),
                        " ".join(["{}={}".format(k, v) for k, v in v.items()])
                    )
                )
        return
