import json
import time

from si_tests import logger
from si_tests.clients import http_client
from si_tests.clients.iam import keycloak_client

LOG = logger.logger


class PrometheusClient(object):
    def __init__(self, host, port, proto, username="admin", password='topsecret',
                 token=None, verify=False):
        self.token = token
        self.token_creation_time = None
        self.refresh_expire = 300
        self.login_expire = 1600
        self.verify = verify
        self.username = username
        self.password = password
        self.keycloak_client = None
        self.url = '{0}://{1}:{2}'.format(proto, host, port)
        self.client = http_client.HttpClient(base_url=self.url)
        if not self.token:
            self.login()

    def refresh_token(func):
        def wrap(self, *args, **kwargs):
            if int(time.time()) > self.token_creation_time + self.login_expire:
                LOG.info("Re-login with OpenID")
                self.login()
            elif int(time.time()) > self.token_creation_time + self.refresh_expire:
                LOG.info("Refresh OpenID token")
                result = self.keycloak_client.refresh_token()
                self.token = result['access_token']
                self.token_creation_time = int(time.time())

            return func(self, *args, **kwargs)

        return wrap

    def compose_header(self):
        header = {"Content-Type": "application/json"}
        if self.token:
            header.update({"Authorization": "Bearer {}".format(self.token)})
        return header

    def login(self):
        raise Exception("Not support")

    def get_targets(self):
        _, resp = self.client.get("/api/v1/targets")
        targets = json.loads(resp)
        return targets["data"]["activeTargets"]

    @refresh_token
    def get_query(self, query, timestamp=None):
        """
        Evaluates an instant query at a single point in time

        Args:
            query: [string] Prometheus expression query string.
            timestamp: [unix_timestamp] Evaluation timestamp

        Returns:

        """
        params = {
            "query": query
        }

        if timestamp is not None:
            params.update({"time": timestamp})

        resp = self.client.get("/api/v1/query", params=params, headers=self.compose_header(), verify=self.verify)

        if not resp.ok:
            raise Exception(resp.text)

        query_result = json.loads(resp)
        if query_result["status"] != "success":
            raise Exception("Failed resp: {}".format(resp))

        if query_result["data"]["resultType"] in ["matrix", "vector"]:
            return query_result["data"]["result"]
        else:
            return query_result

    @refresh_token
    def get_query_range(self, query, start, end, step, timeout=None):
        """
        Evaluates an expression query over a range of time

        Args:
            query: [string] Prometheus expression query string
            start: [unix_timestamp] Start timestamp, inclusive.
            end: [unix_timestamp] End timestamp, inclusive.
            step: [float] Query resolution step width in duration format or float number of seconds.
            timeout: [duration] Evaluation timeout. Optional.

        Returns:

        """
        params = {
            "query": query,
            "start": start,
            "end": end,
            "step": step
        }

        if timeout:
            params.update({"timeout": timeout})

        resp = self.client.get("/api/v1/query_range", params=params, headers=self.compose_header(), verify=self.verify)

        if not resp.ok:
            raise Exception(resp.text)

        query_result = json.loads(resp.text)
        if query_result["status"] != "success":
            raise Exception("Failed resp: {}".format(resp))

        if query_result["data"]["resultType"] in ["matrix", "vector"]:
            return query_result["data"]["result"]
        else:
            return query_result

    def get_svc_probe_duration_seconds(self, namespace, service_name, start, end, step="1"):
        query = f"probe_duration_seconds{{blackbox_probe_module='tls', job='mcc-blackbox', pod='', " \
                f"namespace='{namespace}', service_name='{service_name}'}}"
        return self.get_query_range(query=query, start=start, end=end, step=step)

    def get_svc_probe_success(self, namespace, service_name, start, end, step="1"):
        query = f"probe_success{{blackbox_probe_module='tls', job='mcc-blackbox', pod='', " \
                f"namespace='{namespace}', service_name='{service_name}'}}"
        return self.get_query_range(query=query, start=start, end=end, step=step)


class PrometheusClientOpenid(PrometheusClient):
    def __init__(self, host, port, proto, keycloak_ip,
                 username, password,
                 client_id='sl', token=None, verify=False):
        self.keycloak_ip = keycloak_ip
        self.client_id = client_id  # 'sl 'for StackLight services
        super().__init__(host=host, port=port, proto=proto, username=username,
                         password=password, token=token,
                         verify=verify)

    def login(self):
        """Authorize with username/password and get new token"""

        client = keycloak_client.KeycloakUserClient(self.keycloak_ip,
                                                    self.username,
                                                    self.password,
                                                    client_id=self.client_id,
                                                    realm_name="iam")
        self.keycloak_client = client
        result = client.get_openid_token()
        LOG.debug(f"OpenID data for Prometheus UI:\n{result}")
        self.token = result['id_token']
        self.token_creation_time = int(time.time())
        self.refresh_expire = result['expires_in']
        self.login_expire = result['refresh_expires_in']
        LOG.info("Prometheus UI auth token have been refreshed")
