import exec_helpers
import yaml

from si_tests import settings
from si_tests import logger
from si_tests.deployments.utils import (file_utils, commons)
from si_tests.managers.openstack_manager import OpenStackManager
from si_tests.utils import utils
from si_tests.utils.utils import Provider
from si_tests.managers.openstack_client_manager import OpenStackClientManager
from retry import retry

import pytest
import re
import os.path
LOG = logger.logger

TEMPLATE_PATH = file_utils.join(os.path.dirname(os.path.abspath(__file__)),
                                "../tests/lcm/templates/heat_stack_host.yaml")


@pytest.fixture(scope='session')
def autoset_migration_mode(kaas_manager):
    yield

    ns = kaas_manager.get_namespace(settings.TARGET_NAMESPACE)
    cluster = ns.get_cluster(settings.TARGET_CLUSTER)
    if not re.search(r"mos.?-.*", cluster.clusterrelease_version):
        msg = "Skipping annotation configuration for non-Openstack clusters"
        LOG.info(msg)
        return
    LOG.info("Setting migration mode for different compute types")
    dpdk_machines = [m for m in cluster.get_machines(machine_type="worker") if
                     m.get_k8s_node().data.get('metadata', {}).get('labels', {}).get(
                         'openstack-compute-node-dpdk', '') == 'enabled']
    nova_machines = [m for m in cluster.get_machines(machine_type="worker") if
                     m.get_k8s_node().data.get('metadata', {}).get('labels', {}).get(
                         'openstack-compute-node', '') == 'enabled']
    nova_machines = [m for m in nova_machines if m not in dpdk_machines]
    if dpdk_machines:
        if len(dpdk_machines) == 1:
            LOG.info("Only 1 dpdk compute found. Setting migration mode to SKIP")
            dpdk_machines[0].add_k8s_node_annotations(
                annotations={'openstack.lcm.mirantis.com/instance_migration_mode': 'skip'})
        else:
            LOG.info("More then 1 dpdk compute found. Setting migration mode to LIVE")
            for machine in dpdk_machines:
                machine.add_k8s_node_annotations(
                    annotations={'openstack.lcm.mirantis.com/instance_migration_mode': 'live'})
    else:
        LOG.info("Machines with label openstack-compute-node-dpdk are not found. Skipping")

    if nova_machines:
        if len(nova_machines) == 1:
            LOG.info("Only 1 compute found in nova zone. Setting migration mode to SKIP")
            nova_machines[0].add_k8s_node_annotations(
                annotations={'openstack.lcm.mirantis.com/instance_migration_mode': 'skip'})
        else:
            LOG.info("More then 1 compute found in nova zone. Setting migration mode to LIVE")
            for machine in nova_machines:
                machine.add_k8s_node_annotations(
                    annotations={'openstack.lcm.mirantis.com/instance_migration_mode': 'live'})
    else:
        LOG.info("Machines with label openstack-compute-node are not found")


def verify_stack_workload(openstack_client_manager, stack_name):
    ssh_access_key = openstack_client_manager.stack.show([stack_name])["outputs"][0].get("output_value")
    pkey = utils.get_rsa_key(ssh_access_key)

    @retry(TimeoutError, tries=10, delay=60)
    def _ssh_instance(ip):
        LOG.info(f"Attemp to connect to VM with ip: {ip} in stack {stack_name}")
        auth = exec_helpers.SSHAuth(username="cirros", password='', key=pkey)
        ssh = exec_helpers.SSHClient(
            host=ip, port=22, auth=auth)
        ssh.logger.addHandler(logger.console)
        ssh.sudo_mode = True
        return ssh

    servers = [x["resource_name"] for x in
               openstack_client_manager.stack.resource_list([stack_name, "--filter", "type=OS::Nova::Server"],
                                                            yaml_output=True)]

    for server in servers:
        def _get_fip_only(server):
            server_info = openstack_client_manager.stack.resource_show([stack_name, server])
            network = list(server_info.get('attributes').get('addresses').keys())[0]
            for address in server_info.get('attributes').get('addresses').get(network):
                addr_type = address['OS-EXT-IPS:type']
                if addr_type == 'floating':
                    LOG.info("Verify instance {0} with floating ip {1}".format(
                        server_info.get('attributes').get('OS-EXT-SRV-ATTR:hostname'), address['addr']))
                    return address['addr']

        ip = _get_fip_only(server)
        ssh = _ssh_instance(ip)
        expected_string = utils.gen_random_string(10)
        ssh.check_call('echo {expected_string} > test.txt'.format(expected_string=expected_string))
        res = ssh.check_call('cat test.txt')
        assert res.stdout_brief == expected_string


@pytest.fixture(scope='function')
def create_heat_stacks_before_test(kaas_manager, request):

    cluster_name = settings.TARGET_CLUSTER
    namespace_name = settings.TARGET_NAMESPACE

    ns = kaas_manager.get_namespace(namespace_name)
    child_cluster = ns.get_cluster(cluster_name)
    cr_before = child_cluster.clusterrelease_version

    if "mos" not in cr_before or child_cluster.provider is not Provider.baremetal:
        LOG.info("Skipping 'create_heat_stacks_before_test' fixture because of not suitable cluster")
        yield
        return

    child_kubeconfig_name, child_kubeconfig = child_cluster.get_kubeconfig_from_secret()
    with open('child_conf', 'w') as f:
        f.write(child_kubeconfig)

    openstack_client_manager = OpenStackClientManager(kubeconfig='child_conf')
    os_manager = OpenStackManager(kubeconfig='child_conf')
    # Create test stacks with instances on different hosts
    # Get compute hosts names:

    LOG.info("Get compute hosts names")
    hosts = openstack_client_manager.host.list([])
    LOG.info("Unset quotas for volumes and instances")
    openstack_client_manager.quota.set(["--volumes", "-1", "--instances", "-1", "--cores", "-1",
                                        "--class", "default"])
    for hs in hosts:
        if hs.get('Service') == 'compute' and 'ironic' not in hs.get('Host Name'):
            hostname = hs['Host Name']
            node_labels = child_cluster.k8sclient.nodes.get(name=hostname).read().to_dict().\
                get('metadata', {}).get('labels', {})
            commons.LOG.info(f"Create heat stack before test on {hostname}")
            if node_labels.get('openstack-compute-node-dpdk', '') == 'enabled':
                # Create a flavor with huge pages
                LOG.info("Create a flavor with huge pages")
                flavor_name = "m1.tiny.dpdk-" + hostname
                flavor_params = ["--ram", "1024", "--disk", "5", "--vcpus", "2", "--property", "hw:mem_page_size=2048",
                                 flavor_name]
                keystone_client_pod = [
                    pod for pod in os_manager.get_os_pods() if
                    'keystone-client' in pod.name]
                assert len(keystone_client_pod) > 0, (
                    "No pods found with keystone-client name prefix")
                keystone_client_pod = keystone_client_pod[0]
                flavors_cmd = [
                    '/bin/sh', '-c',
                    'openstack flavor list --all --long -f json']
                flavors = yaml.safe_load(
                    keystone_client_pod.exec(flavors_cmd))
                current_flavor = [flavor for flavor in flavors if flavor.get("Name") == flavor_name]
                if current_flavor:
                    LOG.info(f'Flavor with name: {flavor_name} already exists')
                else:
                    LOG.info(f'Creating new flavor with name: {flavor_name}')
                    openstack_client_manager.flavor.create(flavor_params)

                stack_params = {"host_name": hs['Zone'] + ":" + hostname, "flavor1": "m1.tiny.dpdk-" + hostname,
                                "key_name": "key-" + hostname, "image_name_1": settings.OPENSTACK_CIRROS_IMAGE_NAME}
                openstack_client_manager.create_stack(request, "stack-" + hostname, TEMPLATE_PATH,
                                                      kubeconfig='child_conf', stack_params=stack_params,
                                                      finalizer=False)
                if not child_cluster.tf_enabled():
                    verify_stack_workload(openstack_client_manager, "stack-" + hostname)

            elif node_labels.get('openstack-compute-node', '') == 'enabled':
                stack_params = {"host_name": hs['Zone'] + ":" + hostname, "key_name": "key-" + hostname,
                                "image_name_1": settings.OPENSTACK_CIRROS_IMAGE_NAME}
                openstack_client_manager.create_stack(request, "stack-" + hostname, TEMPLATE_PATH,
                                                      kubeconfig='child_conf', stack_params=stack_params,
                                                      finalizer=False)
                if not child_cluster.tf_enabled():
                    verify_stack_workload(openstack_client_manager, "stack-" + hostname)

    yield

    commons.LOG.info("Check heat stacks after test")
    for hs in hosts:
        if hs.get('Service') == 'compute' and 'ironic' not in hs.get('Host Name'):
            hostname = hs['Host Name']
            node_labels = child_cluster.k8sclient.nodes.get(name=hostname).read().to_dict(). \
                get('metadata', {}).get('labels', {})
            if (node_labels.get('openstack-compute-node', '') == 'enabled' or
                    node_labels.get('openstack-compute-node-dpdk', '') == 'enabled'):
                stack_name = "stack-" + hostname
                try:
                    openstack_client_manager.check_stack(stack_name)
                    if not child_cluster.tf_enabled():
                        verify_stack_workload(openstack_client_manager, stack_name)
                    openstack_client_manager.stack.delete(["stack-" + hostname, "-y", "--wait"])
                except Exception as e:
                    LOG.error(f"Stack {stack_name} verification failed (skipping):\n{e}")
