#!/usr/bin/python
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ansible.errors import AnsibleError, AnsibleFilterError
import random


class MacGenerator:
    def __init__(self, prefix):
        if prefix:
            self.template = [x.strip() for x in prefix.lower().split(':')[:6]]
        else:
            self.template = []
        self.template.extend('*' * (6 - len(self.template)))

    def random(self):
        mac = []
        for x, item in enumerate(self.template):
            if item in ['*', '**']:
                mac.append(random.randint(0, 255))
            else:
                try:
                    num = int(item, 16)
                    if num > 255:
                        raise AnsibleFilterError(
                            f"Bad number in MAC address template "
                            f"'{self.template}', pos {x}.")
                    mac.append(num)
                except ValueError:
                    raise AnsibleFilterError(
                        f"Bad value in MAC address prefix "
                        f"'{self.template}', pos {x}.")

        return tuple(mac)

    def mac2str(self, mac):
        return ':'.join([f'{x:02x}' for x in mac])

    def generate(self, count=1):
        macs = set()
        attempts = 0
        max_attempts = count + 255

        while len(macs) < count:
            macs.add(self.random())
            attempts += 1
            if attempts > max_attempts:
                raise AnsibleError(
                    f"Failed to generate {count} MACs in "
                    f"{attempts} attempts, {len(macs)} succeeded.")

        return [self.mac2str(x) for x in macs]


def random_macs(prefix, count=1, seed=None):
    if seed:
        random.seed(seed)
    macgen = MacGenerator(prefix=prefix)
    return macgen.generate(count)


class FilterModule:
    def filters(self):
        return {
            'random_macs': random_macs
        }
