#    Copyright 2019 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import json
import pytest
from datetime import datetime
from si_tests.utils import packaging_version as version

from si_tests import logger
from si_tests import settings
from si_tests.managers import bootstrap_manager
from si_tests.utils.utils import Provider
from si_tests.managers.kaas_manager import Manager

LOG = logger.logger


def dump_nodes_info(nwll):
    nodes_data_map = {}
    for node in nwll:
        node_name = node["spec"]["nodeName"]
        nwl_name = node["metadata"]["name"]
        nwl_state = node["status"]["state"]
        nwl_resource_version = node["metadata"]["resource_version"]
        nodes_data_map[node_name] = {"workloadlock_name": nwl_name, "workloadlock_state": nwl_state,
                                     "resource_version": nwl_resource_version}
    return nodes_data_map


class TestUpdateMgmCluster():

    @pytest.mark.usefixtures("store_updated_mgmt_cluster_description")
    @pytest.mark.usefixtures("introspect_management_upgrade_objects")
    @pytest.mark.usefixtures("introspect_PRODX_9696_mgmt")
    @pytest.mark.usefixtures("introspect_ceph_mgmt")
    @pytest.mark.usefixtures('introspect_mgmt_lcm_operation_stuck')
    @pytest.mark.usefixtures('introspect_machines_stages_mgmt')
    def test_update_mgmt_cluster(self, kaas_manager: Manager, mgmt_k8s_ip: str):
        try:
            self.update_mgmt_cluster(kaas_manager)
        except Exception as e:
            LOG.error("Management upgrade failed: {}".format(e))
            raise e
        finally:
            LOG.info('Updating bootstrap version')
            bootstrap = bootstrap_manager.BootstrapManager(seed_ip=settings.SEED_STANDALONE_EXTERNAL_IP)
            bootstrap.update_bootstrap_version(mgmt_k8s_ip)

    def update_mgmt_cluster(self, kaas_manager: Manager):
        """Check mgmt is updated

        Scenario:
            1. Check release
        """

        def sanitize_lcmmachines(lcmmachines):
            sanitized_lcmmachines = []
            for lcm in lcmmachines:
                lcm_status = lcm.get('status') or {}
                sanitized_lcmmachines.append({
                    'status': {
                        'addresses': lcm_status.get('addresses'),
                        'state': lcm_status.get('state')
                    }
                })
            return sanitized_lcmmachines

        cluster = kaas_manager.get_mgmt_cluster()
        # Get the current mgmt kaasrelease version until update is not started
        kr_before = kaas_manager.get_mgmt_cluster().get_kaasrelease_version()
        # Get the current mgmt clusterrelease version until update is not started
        cr_before = kaas_manager.get_mgmt_cluster().clusterrelease_version
        LOG.info(f"Management cluster ClusterRelease version is "
                 f"'{cr_before}' , KaaSRelease is '{kr_before}'")

        ucp_tag_in_cr_before = set(
            [x['params']['ucp_tag'] for x in kaas_manager.get_clusterrelease(
                cr_before).data['spec']['machineTypes']['control']
             if 'ucp_tag' in x['params'].keys()])
        LOG.info("Current management cluster ucp version is "
                 "{}".format(ucp_tag_in_cr_before))

        # Collect reboots count for mgmt and region clusters before update
        reboots_mgmt = {}
        reboots_mgmt_after = {}
        reboots_regions = {}
        reboots_regions_after = {}
        reboots_mgmt = cluster.get_nodes_reboots()

        tmp, kernel_versions_before = cluster.get_nodes_kernel_versions()

        lcmms = kaas_manager.api.kaas_lcmmachines.list_raw().to_dict()
        all_managed_lcmm = [lm for lm in lcmms['items']
                            if lm['metadata']['namespace'] != "default"]
        r_clusters = kaas_manager.get_regional_clusters()
        for r_cluster in r_clusters:
            lcmms = r_cluster.k8sclient.kaas_lcmmachines.list_raw().to_dict()
            all_managed_lcmm.extend([lm for lm in lcmms['items']
                                     if lm['metadata']['namespace'] != r_cluster.namespace])
            reboots_regions[r_cluster.name] = r_cluster.get_nodes_reboots()
        # If scheduled update is enabled, wait and check that there is no new release
        # until we match blockUntil time
        if settings.MCC_SCHEDULED_UPGRADES:
            current_mcc_upg_obj = kaas_manager.get_kaas_mcc_upgrade('mcc-upgrade').data
            LOG.info(f"Initial state mccupgrade crd {current_mcc_upg_obj}")
            # get expected time from mccupgrade obj
            exp_time_row = kaas_manager.get_kaas_mcc_upgrade('mcc-upgrade').data['spec']['blockUntil']
            LOG.info(f"Update should not start before {exp_time_row}")
            while datetime.strptime(exp_time_row, '%Y-%m-%dT%H:%M:%SZ') >= datetime.now():
                avail_kaas_releases = kaas_manager.get_kaasrelease_names()
                assert len(avail_kaas_releases) == 1, (
                    f"There is unexpected kaas releases, current list: {avail_kaas_releases}")

        # Wait until KaaS downloads the second kaasrelease
        LOG.info("Waiting for downloading of new  kaasrelease")
        kaas_manager.wait_kaasrelease(
            releases_count=2, timeout=settings.KAAS_UPDATE_MGMT_WAIT_TIMEOUT)

        avail_kaas_releases = kaas_manager.get_kaasrelease_names()
        LOG.info('Available KaaS releases are {}'.format(avail_kaas_releases))

        assert len(avail_kaas_releases) > 1, (
            "There are not new kaas releases, current list: {}".format(
                avail_kaas_releases))

        kaas_r_sorted = str(max(
            [version.parse(x) for x in avail_kaas_releases]))
        LOG.info('Sorted release {}'.format(kaas_r_sorted))
        expected_cr = kaas_manager.get_kaasrelease(kaas_r_sorted).data['spec']['clusterRelease']
        is_patchrelease = cluster.is_patchrelease(expected_cr)

        reboot_required = cluster.update_requires_reboot_mgmt(
            cr_before=cr_before, kaasrelease_version=kaas_r_sorted)
        bm_regions = []
        reboot_required_region = False
        for r_cluster in r_clusters:
            reboot_required_region |= r_cluster.update_requires_reboot_mgmt(
                cr_before=cr_before, kaasrelease_version=kaas_r_sorted)
            if r_cluster.provider is Provider.baremetal:
                bm_regions.append(r_cluster)

        assert not (is_patchrelease and (reboot_required or reboot_required_region)), (
            f"RebootRequired flag is enabled! ClusterRelease {expected_cr} "
            f"is assumed as a patchrelease, where should not be any reboots")

        # let's pass target kaasrelease_version here to fetch
        # target clusterrelease version for mgmt/region clusters
        if reboot_required or reboot_required_region:
            LOG.info("Restart of nodes will be required during update")

        all_managed_lcmm = sanitize_lcmmachines(all_managed_lcmm)

        kr_data = kaas_manager.get_kaasrelease(kaas_r_sorted).data
        LOG.debug("Release_details {}".format(kr_data))
        kaas_manager.wait_kaasrelease_label(kaas_r_sorted, timeout=settings.KAAS_UPDATE_MGMT_ACTIVE_WAIT_TIMEOUT)
        formated_kr = kr_data['spec']['version'].replace('.', '-')
        assert formated_kr == kaas_r_sorted.replace('kaas-', '')
        mgmt_cluster = kaas_manager.get_mgmt_cluster().spec['providerSpec']
        assert kaas_r_sorted == mgmt_cluster['value']['kaas']['release']
        LOG.info('MGMT spec {}'.format(mgmt_cluster['value']['release']))

        cluster_releases = kaas_manager.get_clusterrelease_names()
        spec_cr = kr_data['spec']['clusterRelease']
        LOG.info('Cluster version in new KaaSRelease: {}'.format(spec_cr))
        assert spec_cr in cluster_releases
        assert spec_cr == mgmt_cluster['value']['release'], (
            "Cluster version in KaasRelease and Mgmt "
            "Cluster differs, that's wrong")

        cluster.check.check_cluster_release(spec_cr)
        # TODO: we need to check regional child cluster updates too
        if cr_before != expected_cr:
            cluster.check.check_update_finished(check_managed_clusters=True)

        if cluster.provider is Provider.baremetal and cr_before != expected_cr:
            kaascephclusters = cluster.get_kaascephclusters()
            num_clusters = len(kaascephclusters)
            assert num_clusters == 0, (f"Unexpectedly {num_clusters} kaascephclusters found "
                                       f"with spec.k8sCluster '{self.namespace}/{self.name}' : "
                                       f"{kaascephclusters}")

        ucp_tag_after = set(
            [x['params']['ucp_tag']
             for x in kaas_manager.get_clusterrelease(
                spec_cr).data['spec']['machineTypes']['control']
             if 'ucp_tag' in x['params'].keys()])
        if ucp_tag_in_cr_before != ucp_tag_after:
            LOG.info("Management cluster new ucp version will be "
                     "{}".format(ucp_tag_after))

        cluster.check.check_cluster_nodes()
        cluster.check.check_cluster_readiness(timeout=3600)
        cluster.check.check_helmbundles()
        cluster.check.check_k8s_nodes()
        cluster.check.check_persistent_volumes_mounts()
        cluster.check.check_k8s_pods()
        cluster.check.check_upgrade_stage_success()
        if cluster.provider is Provider.baremetal:
            LOG.info("Check repository url for management cluster")
            cluster.check.check_repository_url()
            cluster.check.check_actual_expected_kernel_versions()
        if bm_regions:
            for r_cluster in bm_regions:
                if version.parse(mgmt_cluster['value']['kaas']['release']) < version.parse("kaas-2-20-0-rc"):
                    r_cluster.check.check_ceph_pvc()
                LOG.info("Check repository url for regional cluster")
                r_cluster.check.check_repository_url()
                r_cluster.check.check_actual_expected_kernel_versions()

        # Check/wait for correct docker service replicas in cluster
        cluster.check.check_actual_expected_docker_services()
        cluster.check.check_actual_expected_pods(timeout=600)
        # TODO enable after verification on all providers
        # Related blocker PRODX-19064
        # cluster.check.check_actual_expected_rolebindings()

        # Check the changes in reboots count on mgmt cluster after upgrade
        reboots_mgmt_after = cluster.get_nodes_reboots()
        if not reboot_required:
            assert reboots_mgmt == reboots_mgmt_after, \
                f"Unexpected number of reboots: expected {reboots_mgmt}" \
                f"actual: {reboots_mgmt_after}. Reboot wasn't expected"
        else:
            for k, v in reboots_mgmt.items():
                assert int(v) < int(reboots_mgmt_after[k]), \
                    f"Unexpected number of reboots for {k}: " \
                    f"expected {v}, actual: {reboots_mgmt_after[k]}. " \
                    f"At least 1 reboot was expected\n" \
                    f"All reboots before update: {reboots_mgmt},\n " \
                    f"after: {reboots_mgmt_after}"

        # Check that all workloadlock objects are changed their state
        if cluster.provider is Provider.baremetal and cr_before != expected_cr and reboot_required:
            # Refresh nodeworkloadlock dict
            nwll = [nwl.data for nwl in cluster.get_nodeworkloadlocks()]
            # NodeWorkloadLocks should be empty on management cluster as there is no Ceph cluster
            assert len(nwll) == 0, f"NodeWorkLoadLocks list is not empty in mgm cluster: {cluster.name}"

        lcmms = \
            kaas_manager.api.kaas_lcmmachines.list_raw().to_dict()
        all_managed_lcmm_new = [lm for lm in lcmms['items']
                                if lm['metadata']['namespace'] != "default"]

        if reboot_required and cluster.provider is Provider.baremetal:
            operating_system, \
                kernel_versions_after = cluster.get_nodes_kernel_versions()
            for k, v in kernel_versions_before.items():
                assert version.parse(v) != version.parse(
                    kernel_versions_after[k]), \
                    f"Unexpected kernel version on machine {k}" \
                    f"({operating_system[k]}) after upgrade: " \
                    f"before: {v}, after: {kernel_versions_after[k]}"
            LOG.info(f"Kernel version updated successfully. From version {v}, "
                     f"to {kernel_versions_after[k]}")
        else:
            LOG.info("Skip checking updated kernel version, "
                     f"as it is not expected or provider {cluster.provider} "
                     "is skipped")

        for r_cluster in r_clusters:
            # get regional clusters and check CR there
            r_cr = r_cluster.spec['providerSpec']['value']['release']
            LOG.info("Current regional cr is {}".format(r_cr))
            assert spec_cr in r_cr, ("Cluster version in regional cluster "
                                     "differs from latest kaas release, "
                                     "that's wrong")
            r_cluster.check.check_cluster_release(spec_cr)

            r_cluster.check.check_cluster_nodes()
            r_cluster.check.check_cluster_readiness()
            r_cluster.check.check_helmbundles()
            r_cluster.check.check_k8s_nodes()
            r_cluster.check.check_persistent_volumes_mounts()
            r_cluster.check.check_k8s_pods()
            r_cluster.check.check_upgrade_stage_success()

            # Check/wait for correct docker service replicas in cluster
            r_cluster.check.check_actual_expected_docker_services()
            r_cluster.check.check_actual_expected_pods(timeout=600)
            # TODO enable after verification on all providers
            # Related blocker PRODX-19064
            # r_cluster.check.check_actual_expected_rolebindings()

            if r_cluster.provider is Provider.baremetal:
                LOG.info("Check repository url for regional cluster")
                r_cluster.check.check_repository_url()

            lcmms = \
                r_cluster.k8sclient.kaas_lcmmachines.list_raw().to_dict()
            all_managed_lcmm_new.extend(
                [lm for lm in lcmms['items']
                 if lm['metadata']['namespace'] != r_cluster.namespace]
            )
            reboots_regions_after[r_cluster.name] = r_cluster.get_nodes_reboots()

        LOG.info("Check child clusters")
        child_clusters = kaas_manager.get_child_clusters()
        for child in child_clusters:
            LOG.info(f" - check {child.name} cluster")
            child.check.check_cluster_readiness()
            if child.provider is Provider.baremetal:
                LOG.info("Check repository url for child cluster")
                child.check.check_repository_url()

        all_managed_lcmm_new = sanitize_lcmmachines(all_managed_lcmm_new)

        LOG.debug(f"All managed lcmmachines before update: "
                  f"{json.dumps(all_managed_lcmm, indent=4, default=str)}")
        LOG.debug(f"All managed lcmmachines after update: "
                  f"{json.dumps(all_managed_lcmm_new, indent=4, default=str)}")

        # Check the changes in reboots count on region clusters after upgrade
        if not reboot_required_region:
            assert reboots_regions == reboots_regions_after, \
                f"Unexpected number of reboots: expected {reboots_regions}" \
                f"actual: {reboots_regions_after}. Reboot wasn't expected"
        else:
            for clust, data in reboots_regions.items():
                for k, v in data.items():
                    assert int(v) < int(reboots_regions_after[clust][k]), \
                        f"Unexpected number of reboots for cluster {clust} " \
                        f"node {k}: expected {v}, " \
                        f"actual: {reboots_regions_after[k]}. " \
                        f"At least 1 reboot was expected\n" \
                        f"All reboots before update: {reboots_regions},\n " \
                        f"after: {reboots_regions_after}"

        assert all_managed_lcmm_new == all_managed_lcmm, \
            "Error: lcmmachines of managed cluster have changed"

        LOG.info("Check clusterRelease objects are existed for every release "
                 "announced as supported in kaasRelease object")
        cluster.check.compare_supported_and_existed_clusterreleases()
        if cluster.provider is Provider.baremetal:
            LOG.info("Check actual/expected reboot-required status for all machines")
            cluster.check.check_expected_actual_reboot_required_status()

        if is_patchrelease:
            cluster.check.check_upgraded_machines_not_contain_cordon_drain_stages()
